from numpy import *
from numpy.typing import *
from numpy.linalg import norm

def cropDcf(arrDcf:NDArray, arrK:NDArray, nPix:int, nNyq:int=2) -> NDArray: 
    '''
    Null the outmost DCF points, useful for some methods including Voronoi, iterative methods.
    
    Args:
        arrDcf: array of DCF, shape: `[Nk,]`
        arrK: array of trajectory, shape: `[Nk,Nd]`, range: `[-0.5,0.5]`
        nPix: number of pixel, for calc the Nyquist Interval
        nNqy: DCF data within how long distance to the edge will be removed
    
    Returns:
        Cropped DCF.
    '''
    arrRho = norm(arrK, axis=-1)
    arrDcf[arrRho>0.5-nNyq/nPix] = 0
    return arrDcf


def normDcf(arrDcf:NDArray, nAx:int) -> NDArray:
    """
    Deprecated. Make the sum of DCF equivalent to the area of the inner circle of k-space in 2D, or a inner ball of k-space in 3D. Inspired by Voronoi method.

    Args:
        arrDcf: Input DCF.
        nAx: Number of dimensions.

    Returns:
        Normalized DCF.
    """
    arrDcf = arrDcf/abs(arrDcf).sum()
    if nAx == 2: arrDcf *= pi/4
    if nAx == 3: arrDcf *= pi/6
    return arrDcf

def normImg(arrData:NDArray, method:str="mean0_std1", mskFov:NDArray|None=None) -> NDArray:
    """
    Normalize a image using different methods.

    Args:
        arrData: Input image.
        method: Normalization strategy. Support values:
            "mean0_std1": Zero mean and unit standard deviation.
            "mean1_std1": Unit standard deviation with mean shifted by 1.
            "mean": Unit mean.
            "std": Unit deviation.
            "max": Unit max magnitude.
            "ene": Unit L2 norm (energy).
        mskFov: Optional mask for the Field of View. If provided, statistics (mean, std, etc.) are calculated only using pixels within the mask.

    Returns:
        Normalized image.
    """
    arrData = arrData.copy()
    
    if mskFov is None:
        _arrData = arrData
    else:
        _arrData = arrData[mskFov]
    vmean = _arrData.mean()
    vstd = _arrData.std()
    vmax = abs(_arrData).max()
    vene = norm(_arrData.flatten())
        
    if method=="mean0_std1":
        arrData -= vmean
        arrData /= vstd
    elif method=="mean1_std1":
        arrData -= vmean
        arrData /= vstd
        arrData += vmean/abs(vmean)
    elif method=="mean":
        arrData /= abs(vmean)
    elif method=="std":
        arrData /= vstd
    elif method=="max":
        arrData /= vmax
    elif method=="ene":
        arrData /= vene
    else:
        raise NotImplementedError("")
    return arrData


from numpy import *
from numpy.linalg import norm
from matplotlib.pyplot import *
from scipy.io import loadmat
import mrarbdcf as mad

nPix = 256
pathRes = "/mnt/d/LProject/DcfBenchmark/resource/"
pathFig = "dev/figure/"
# nameM0 = "arrM0_3d.npy"
nameM0 = "imgRec.npy"
# nameM0 = "1b197efe-9865-43be-ac24-f237c380513e.npy"
# nameM0 = "2588bfa8-0c97-478c-aa5a-487cc88a590d.npy"

# show metrics table
for sTraj in ["VdSpiral", "Rosette", "Yarnball", "Cones"]:
    dicResult = dict()
    for sMethod in ["Baseline", "Proposed"]:
        dicResult[sMethod] = load(f"{pathRes}{sTraj}_{sMethod}.npz")
    print(f"{dicResult["Baseline"]["metNrmse"]:.3f} & {dicResult["Proposed"]["metNrmse"]:.3f} & {dicResult["Baseline"]["metSsim"]:.3f} & {dicResult["Proposed"]["metSsim"]:.3f}")
print()
for sTraj in ["VdSpiral", "Rosette", "Yarnball", "Cones"]:
    dicResult = dict()
    for sMethod in ["Baseline", "Proposed"]:
        dicResult[sMethod] = load(f"{pathRes}{sTraj}_{sMethod}.npz")
    print(f"{dicResult["Baseline"]["tExe"]:.3f} & {dicResult["Proposed"]["tExe"]:.3f} & {around(dicResult["Baseline"]["tExe"]/dicResult["Proposed"]["tExe"]).astype(int)}")
print()

# exit() # test

# load trajectory
sTraj = "Yarnball"; nAx = 3
dicTraj = loadmat(f"{pathRes}arrK_{sTraj}.mat")
arrK = asarray(dicTraj["arrK"]).astype(float32)
arrNRO = asarray(dicTraj["arrNRO"]).astype(int).squeeze()
sTraj = str(dicTraj["sTraj"].item())
arrI0 = zeros((arrNRO.size+1,), dtype=int)
arrI0[1:] = cumsum(arrNRO)
lstArrK = [arrK[arrI0[i]:arrI0[i+1]] for i in range(arrI0.size-1)]

# load results
arrM0 = load(f"{pathRes}{nameM0}")
arrM0 = asarray(arrM0)
mskFov = load(f"{pathRes}mskFov_{nAx}d.npy")
mskFov = asarray(mskFov)
mskFov = ones_like(mskFov) # test

dicResult = dict()
for sMethod in ["Baseline", "Proposed"]:
    dicResult[sMethod] = load(f"{pathRes}{sTraj}_{sMethod}.npz")

# # see Rosette results
# figure()
# arrM0 = asarray(arrM0)[nPix//2,...]
# subplot(221)
# arrM0_Bl = load(f"{pathRes[:-1]}/Rosette_Baseline.npz")["arrM0Rec"]
# imshow(abs(arrM0_Bl), cmap="gray")
# clim(0,1)
# subplot(222)
# arrM0_Pp = load(f"{pathRes[:-1]}/Rosette_Proposed.npz")["arrM0Rec"]
# imshow(abs(arrM0_Pp), cmap="gray")
# clim(0,1)
# subplot(223)
# arrErr_Bl = arrM0 - arrM0_Bl
# imshow(abs(arrErr_Bl), cmap="viridis")
# colorbar()
# subplot(224)
# arrErr_Pp = arrM0 - arrM0_Pp
# imshow(abs(arrErr_Pp), cmap="viridis")
# colorbar()
# show()
# exit()

# plot
lstTitle = [*"ABCDEFGHIJKL"]
wFPage = 4.77
dpi = 300

# DCF
# iDcfMax = argmax(dicResult["Proposed"]["arrDcf"])
# iK0 = arrI0[arrI0<=iDcfMax][-1]
# iK1 = arrI0[arrI0>iDcfMax][0]
iK0 = arrI0[0]
iK1 = arrI0[1]

figure(dpi=150, figsize=(wFPage,wFPage*2/3))
subplot(111)
for sMethod in ["Baseline", "Proposed"]:
    arrDcf = dicResult[sMethod]["arrDcf"]
    plot(abs(arrDcf[iK0:iK1]), ".-", markersize=4, linewidth=1, label=sMethod)
arrDcf = dicResult["Baseline"]["arrDcf"][iK0:iK1]
ylim(-0.1*abs(arrDcf).max(), 1.1*abs(arrDcf).max())
xlabel("index [a.u.]")
ylabel("DCF [a.u.]")
legend(loc=2)
grid(True, "both", "both")
subplots_adjust(0.15,0.15,0.99,0.9)
savefig(f"{pathFig}DCF.png", dpi=dpi)

# Phantom
fig = figure(dpi=150, figsize=(wFPage,wFPage*2/3))
wFig = 20
lstAx = [None for _ in range(8)]
gsAx = fig.add_gridspec(2, 3*wFig+1)
lstAx[0] = fig.add_subplot(gsAx[0,0*wFig:1*wFig])
lstAx[1] = fig.add_subplot(gsAx[0,1*wFig:2*wFig])
lstAx[2] = fig.add_subplot(gsAx[0,2*wFig:3*wFig])
lstAx[3] = fig.add_subplot(gsAx[0,3*wFig])
lstAx[4] = fig.add_subplot(gsAx[1,0*wFig:1*wFig])
lstAx[5] = fig.add_subplot(gsAx[1,1*wFig:2*wFig])
lstAx[6] = fig.add_subplot(gsAx[1,2*wFig:3*wFig])
lstAx[7] = fig.add_subplot(gsAx[1,3*wFig])

vmean = abs(arrM0).mean()
vstd = abs(arrM0).std()
vmin = 0 # vmean - 1*vstd
vmax = 1 # vmean + 3*vstd

lstAx[0].imshow(abs(arrM0[nPix//2,:,:]), origin="lower", cmap="gray", vmin=vmin, vmax=vmax)
lstAx[1].imshow(abs(arrM0[:,nPix//2,:]), origin="lower", cmap="gray", vmin=vmin, vmax=vmax)
m = lstAx[2].imshow(abs(arrM0[:,:,nPix//2]), origin="lower", cmap="gray", vmin=vmin, vmax=vmax)
fig.colorbar(m, cax=lstAx[3])

lstAx[4].imshow(angle(arrM0[nPix//2,:,:]), origin="lower", cmap="hsv", vmin=-pi, vmax=pi)
lstAx[5].imshow(angle(arrM0[:,nPix//2,:]), origin="lower", cmap="hsv", vmin=-pi, vmax=pi)
m = lstAx[6].imshow(angle(arrM0[:,:,nPix//2]), origin="lower", cmap="hsv", vmin=-pi, vmax=pi)
fig.colorbar(m, cax=lstAx[7])
lstAx[7].set_yticks([-pi,pi], [r"-$\pi$",r"$\pi$"])

lstIAx = [0,1,2,4,5,6]
for i in range(6):
    iAx = lstIAx[i]
    lstAx[iAx].set_title(f"({lstTitle[i]})", loc="left")
    lstAx[iAx].set_xticklabels([])
    lstAx[iAx].set_yticklabels([])
    lstAx[iAx].set_axis_off()

subplots_adjust(0.01,0.05,0.9,0.9, wspace=0.01, hspace=0.2)
savefig(f"{pathFig}Phantom.png", dpi=dpi)

# show()
# exit()

# recon result
ovPsf = 10
fig = figure(dpi=150, figsize=(wFPage,wFPage*4/3))
gsFig = fig.add_gridspec(2,1)
lstMethod = ["Baseline", "Proposed"]
for i in range(2):
    sMethod = lstMethod[i]
    arrImg = dicResult[sMethod]["arrM0Rec"]
    arrErr = arrImg - arrM0 * norm(arrImg.ravel())/norm(arrM0.ravel())
    
    arrImg *= mskFov
    arrErr *= mskFov
    
    subfig = fig.add_subfigure(gsFig[i])

    wFig = 20
    gsAx = subfig.add_gridspec(2, 3*wFig+1)
    lstAx = [None for _ in range(8)]
    lstAx[0] = subfig.add_subplot(gsAx[0,0*wFig:1*wFig])
    lstAx[1] = subfig.add_subplot(gsAx[0,1*wFig:2*wFig])
    lstAx[2] = subfig.add_subplot(gsAx[0,2*wFig:3*wFig])
    lstAx[3] = subfig.add_subplot(gsAx[0,3*wFig])
    lstAx[4] = subfig.add_subplot(gsAx[1,0*wFig:1*wFig])
    lstAx[5] = subfig.add_subplot(gsAx[1,1*wFig:2*wFig])
    lstAx[6] = subfig.add_subplot(gsAx[1,2*wFig:3*wFig])
    lstAx[7] = subfig.add_subplot(gsAx[1,3*wFig])

    vmean = abs(arrImg).mean()
    vstd = abs(arrImg).std()
    vmin = 0 # vmean - 1*vstd
    vmax = 1 # vmean + 3*vstd

    lstAx[0].imshow(abs(arrImg[nPix//2,:,:]), origin="lower", cmap="gray", vmin=vmin, vmax=vmax)
    lstAx[1].imshow(abs(arrImg[:,nPix//2,:]), origin="lower", cmap="gray", vmin=vmin, vmax=vmax)
    m = lstAx[2].imshow(abs(arrImg[:,:,nPix//2]), origin="lower", cmap="gray", vmin=vmin, vmax=vmax)
    subfig.colorbar(m, cax=lstAx[3])
    
    vmean = abs(arrErr).mean()
    vstd = abs(arrErr).std()
    vscale = 4 # floor(1/(vmean+3*vstd))
    vmin = 0
    vmax = 1/vscale

    lstAx[4].imshow(abs(arrErr[nPix//2,:,:]), origin="lower", cmap="viridis", vmin=vmin, vmax=vmax)
    lstAx[5].imshow(abs(arrErr[:,nPix//2,:]), origin="lower", cmap="viridis", vmin=vmin, vmax=vmax)
    m = lstAx[6].imshow(abs(arrErr[:,:,nPix//2]), origin="lower", cmap="viridis", vmin=vmin, vmax=vmax)
    subfig.colorbar(m, cax=lstAx[7])
    
    for j in [4,5,6]:
        lstAx[j].text(0,0,rf"{vscale:.0f}$\times$", color="#FFFFFF", horizontalalignment='left', verticalalignment='bottom')
    
    lstIAx = [0,1,2,4,5,6]
    for j in range(6):
        iAx = lstIAx[j]
        lstAx[iAx].set_title(f"({lstTitle[j+i*6]})", loc="left")
        lstAx[iAx].set_xticklabels([])
        lstAx[iAx].set_yticklabels([])
        lstAx[iAx].tick_params("both", length=0)

    subplots_adjust(0.01,0.05,0.9,0.9, wspace=0.01, hspace=0.2)

fig.savefig(f"{pathFig}Recon.png", dpi=dpi)

# FWHM
for sMethod in ["Baseline", "Proposed"]:
    arrPsf = dicResult[sMethod]["arrPsf"]
    FWHM = zeros((nAx,), dtype=float32)
    FWHM[0] = count_nonzero(arrPsf[:,nPix,nPix] > abs(arrPsf).max()/2) / ovPsf
    FWHM[1] = count_nonzero(arrPsf[nPix,:,nPix] > abs(arrPsf).max()/2) / ovPsf
    FWHM[2] = count_nonzero(arrPsf[nPix,nPix,:] > abs(arrPsf).max()/2) / ovPsf
    print(f"mean FWHM: {FWHM.mean():.3f}")

# PSF
fig = figure(figsize=(wFPage,wFPage*2/3), dpi=150)
wFig = 20
gsAx = fig.add_gridspec(2,3*wFig+1)
lstAx = [None for _ in range(8)]
lstAx[0] = fig.add_subplot(gsAx[0,0:1*wFig], projection='3d')
lstAx[1] = fig.add_subplot(gsAx[0,1*wFig:2*wFig], projection='3d')
lstAx[2] = fig.add_subplot(gsAx[0,2*wFig:3*wFig], projection='3d')
lstAx[3] = fig.add_subplot(gsAx[1,0:1*wFig], projection='3d')
lstAx[4] = fig.add_subplot(gsAx[1,1*wFig:2*wFig], projection='3d')
lstAx[5] = fig.add_subplot(gsAx[1,2*wFig:3*wFig], projection='3d')
lstAx[6] = fig.add_subplot(gsAx[:,3*wFig])

X, Y = meshgrid(arange(2*nPix), arange(2*nPix))

for i in range(7):
    if i!=6:
        if i//4==0: arrPsf = dicResult["Baseline"]["arrPsf"]
        else: arrPsf = dicResult["Proposed"]["arrPsf"]
        if i%4==0: Z = abs(arrPsf[:,:,nPix])
        if i%4==1: Z = abs(arrPsf[:,nPix,:])
        if i%4==2: Z = abs(arrPsf[nPix,:,:])
        Z /= abs(arrPsf).max()
        m = lstAx[i].plot_surface(X, Y, Z, cmap='viridis', edgecolor='none', vmin=1e1, vmax=1e9, norm="log")
        lstAx[i].set_title(f"({lstTitle[i]})", loc="left")
        lstAx[i].set_axis_off()
    elif i==6:
        fig.colorbar(m, cax=lstAx[i])

fig.subplots_adjust(0.01,0.1,0.9,0.8, wspace=0.01)
fig.savefig(f"{pathFig}PSF.png", dpi=dpi)

show()


from numpy import *
import skimage as ski
from .Type import Part
from .Utility import Enum2M0, Enum2T1, Enum2T2, Enum2Om, genPhMap, genB0Map, genCsm


# update masks
def _updateMask_Dynamic(
    z:int|float, nPix:int,
    # mask to be updated
    mskFatOt:ndarray,
    mskFatIn:ndarray,
    mskMyoOt:ndarray,
    mskMyoIn:ndarray,
    # motion parameter
    ampRes:float|int,
    ampCar:float|int,
) -> None:
    # masks
    mskFatOt.fill(0)
    mskFatIn.fill(0)
    mskMyoOt.fill(0)
    mskMyoIn.fill(0)

    # draw body parts
    # Outer border of fat (expanding and contracting with breathing)
    y = 0 # -(1/2-abs(z/nPix))*nPix*ampRes
    x = 0
    rY = nPix*400e-3 + nPix*ampRes
    rX = nPix*400e-3 - 0.5*nPix*ampRes
    rZ = nPix*480e-3
    rhs = 1 - (z/rZ)**2
    # print(f"rY: {rY:.2f}, rX: {rX:.2f}")
    # print(f"rhs: {rhs:.4f}, rY: {sqrt(rY**2*rhs):.2f}, rX: {sqrt(rX**2*rhs):.2f}")
    if rhs >= 0:
        tupPtFatOt = ski.draw.ellipse(y+nPix//2, x+nPix//2, sqrt(rY**2*rhs), sqrt(rX**2*rhs), (nPix,nPix), pi*0e-2)
        mskFatOt[tupPtFatOt] = 1
    
    # Inner border of fat
    y = 0 # -(1/2-abs(z/nPix))*nPix*ampRes
    x = 0
    rY = nPix*380e-3 + nPix*ampRes
    rX = nPix*380e-3 - 0.5*nPix*ampRes
    rZ = nPix*450e-3
    rhs = 1 - (z/rZ)**2
    if rhs >= 0:
        tupPtFatIn = ski.draw.ellipse(y+nPix//2, x+nPix//2, sqrt(rY**2*rhs), sqrt(rX**2*rhs), (nPix,nPix), pi*4e-2)
        mskFatIn[tupPtFatIn] = 1

    # draw heart
    # Outer ellipse
    y = 0 # -(1/2-abs(z/nPix))*nPix*ampRes
    x = 0
    rY = nPix*100e-3 + nPix*ampCar
    rX = nPix*120e-3 + nPix*ampCar
    rZ = rY
    rhs = 1 - (z/rZ)**2
    if rhs >= 0:
        tupPtMyoOt = ski.draw.ellipse(y+nPix//2, x+nPix//2, sqrt(rY**2*rhs), sqrt(rX**2*rhs), (nPix,nPix))
        mskMyoOt[tupPtMyoOt] = 1

    # Inner ellipse
    y = 0 # -(1/2-abs(z/nPix))*nPix*ampRes
    x = -nPix*20e-3
    rY = nPix*60e-3 + nPix*2*ampCar
    rX = nPix*60e-3 + nPix*2*ampCar
    rZ = rY
    rhs = 1 - (z/rZ)**2
    if rhs > 0:
        tupPtMyoIn = ski.draw.ellipse(y+nPix//2, x+nPix//2, sqrt(rY**2*rhs), sqrt(rX**2*rhs), (nPix,nPix))
        mskMyoIn[tupPtMyoIn] = 1

def _updateMask_Fixed(
    z:int|float, nPix:int,
    lstMskEl:list[ndarray],
) -> None:
    # masks
    for msk in lstMskEl: msk.fill(0)

    # draw eln
    arrY, arrX = meshgrid\
    (
        arange(-nPix//2,nPix//2),
        arange(-nPix//2,nPix//2),
        indexing="ij"
    )
    arrYX = array([arrY,arrX]).transpose(1,2,0)

    nEl = len(lstMskEl)
    r0 = (nPix*4e-1)/(nEl+1)/2; r0 *= 3/4
    arrOzy = array([
        [0, -nPix*0.20],
        [0, nPix*0.20],
        [-nPix*0.25, 0],
        [nPix*0.25, 0],
    ]); arrOzy = around(arrOzy)
    for Oz, Oy in arrOzy:
        arrOyx = array([
            Oy*ones([nEl]),
            linspace(-nPix*2e-1, nPix*2e-1, nEl+2, 1)[1:-1],
        ]).T; arrOyx = around(arrOyx)
        
        for iM in range(nEl):
            Oy, Ox = arrOyx[iM,:]
            r = r0*(1 - abs(Ox-nPix*2e-1)/(nPix*4e-1))
            lstMskEl[iM][sum((arrYX-arrOyx[iM,:])**2,axis=-1) < r**2 - (z-Oz)**2] = 1

def genPhan\
(
    nDim:int=2, nPix:int=256,
    # motion parameter
    arrAmp:ndarray|None=None,
    # whether to return M0 map, T1 map, T2 map, etc, or original enum phantom
    rtM0:bool=True, rtT1:bool=False, rtT2:bool=False, rtOm:bool=False, rtPhan:bool=False, 
    # number of additional ellipsoids for resolution test
    nEl:int=5,
) -> dict|ndarray[uint8]:
    assert nDim==2 or nDim==3
    if nDim==2:
        arrZ = array([0])
        if arrAmp is None: arrAmp = array([[60e-3,0e-3]])
    if nDim==3:
        arrZ = arange(-nPix//2,nPix//2)
        if arrAmp is None: arrAmp = array([[0e-3,0e-3]])
    nT = arrAmp.shape[0]
    nZ = arrZ.size
    # image array
    arrPhan = zeros([nT,nZ,nPix,nPix], dtype=uint8)

    # masks
    mskFatOt = zeros([nPix,nPix], dtype=bool)
    mskFatIn = zeros([nPix,nPix], dtype=bool)
    mskMyoOt = zeros([nPix,nPix], dtype=bool)
    mskMyoIn = zeros([nPix,nPix], dtype=bool)
    lstMskEl = [zeros([nPix,nPix], dtype=bool) for _ in range(nEl)]

    # generate image
    arrPhan.fill(0)
    for iZ in range(nZ):
        z = arrZ[iZ]
        _updateMask_Fixed \
        (
            z, nPix,
            lstMskEl
        )
        for iT in range(nT):

            _updateMask_Dynamic \
            (
                z, nPix,
                mskFatOt, mskFatIn, mskMyoOt, mskMyoIn,
                arrAmp[iT,0], arrAmp[iT,1]
            )
        
            # fill fat
            arrPhan[iT][iZ][mskFatOt & ~mskFatIn] = Part.Fat.value
            # fill body
            arrPhan[iT][iZ][mskFatIn & ~mskMyoOt] = Part.Body.value
            # fill myocardium
            arrPhan[iT][iZ][mskMyoOt & ~mskMyoIn] = Part.Myo.value
            # fill blood pool
            arrPhan[iT][iZ][mskMyoIn] = Part.Blood.value
            # fill ellipsoid_n
            for msk in lstMskEl:
                arrPhan[iT][iZ][msk] = Part.Other.value

    # dictionary to be returned
    dic = dict()
    if rtPhan: dic["Phan"] = arrPhan
    if rtM0: dic["M0"] = Enum2M0(arrPhan)*genPhMap(nDim=nDim, nPix=nPix)
    if rtT1: dic["T1"] = Enum2T1(arrPhan)
    if rtT2: dic["T2"] = Enum2T2(arrPhan)
    if rtOm: dic["Om"] = Enum2Om(arrPhan) + genB0Map(nDim=nDim, nPix=nPix)

    return dic if len(dic) != 0 else arrPhan


py_sum = sum

# import library
import numpy
from numpy.typing import NDArray
import finufft
from time import time
from itertools import product
from psutil import cpu_count

# gpu library
try:
    import cufinufft, cupy
    useCuda = bool(1)
except ImportError:
    useCuda = bool(0)
xp = cupy if useCuda else numpy

pi = numpy.pi

# interface function
nStep = 2 # 1
def setNumStep(x:int):
    """
    While this method can converge in 1 step in most cases, it's possible to set the number of steps via this interface.

    Args:
        x: Number of steps to be set.
    """
    global nStep
    nStep = int(x)
    
fDbgInfo = False
def setDbgInfo(x:bool):
    """
    Set whether to print debug input. Default to False.

    Args:
        x: True for to print. False for not to.
    """
    global fDbgInfo
    fDbgInfo = bool(x)
    
enInputCheck = True
def setInputCheck(x:bool):
    """
    Set whether to check the input type. Recommend and default to True. Can be set to False for benchmark reason.

    Args:
        x: True for to use. False for not to.
    """
    global enInputCheck
    enInputCheck = bool(x)
    
def setUseCuda(x:bool):
    """
    Set whether to use cuda for acceleration.
    By default cuda would be used if cufinufft and cupy are both installed, this function is to let you turn off it.

    Args:
        x: True for to use. False for not to.
    """
    global useCuda, xp
    useCuda = bool(x)
    xp = cupy if useCuda else numpy
    
def _getArrKArrI0(lstArrK:list[NDArray]) -> tuple[NDArray,NDArray]:
    _xp = numpy if isinstance(lstArrK[0], numpy.ndarray) else cupy
    arrK = _xp.concatenate(lstArrK, axis=0).astype(_xp.float32)
    arrNRO = _xp.array([_.shape[0] for _ in lstArrK])
    arrI0 = _xp.zeros((len(arrNRO) + 1,), dtype=int)
    arrI0[1:] = _xp.cumsum(arrNRO)
    return arrK, arrI0
    
def sovDcf(nPix:int, lstArrK:list[NDArray], sWind:str="poly", pShape:float=None) -> list[NDArray]:
    """
    Solve density compensation function.
    
    Args:
        nPix: Designed number of pixels of the trajectory.
        lstArrK: List of trajectorys, element shape [nK,nAx].
        sWind: Window function type, can be "poly", "cos", "es".
        pShape: Window function shape parameter.
        
    Returns:
        list of density compensation function, w.r.t. lstArrK.
    """
    arrK, arrI0 = _getArrKArrI0(lstArrK)
    arrDcf = calDcf(nPix, arrK, arrI0, sWind, pShape)
    lstArrDcf = [arrDcf[i0:i1] for i0,i1 in zip(arrI0[:-1],arrI0[1:])]
    lstArrDcf.append(arrDcf[arrI0[-1]:])
    return lstArrDcf
   
def calDcf(nPix:int, arrK:NDArray, arrI0:NDArray|None=None, sWind:str="poly", pShape:float=None) -> NDArray:
    """
    Deprecated. Calculate density compensation function.
        
    Args:
        nPix: Designed number of pixels of the trajectory.
        arrK: Concatenated trajectories, shape [nK,nAx].
        arrI0: Start index of each interleaves, shape [nIntLea,].
        sWind: Window function type, can be "poly", "cos", "es".
        pShape: Window function shape parameter.
        
    Returns:
        Density compensation function, shape [nK,].
    """
    t0 = time()
    isInputNumpy = isinstance(arrK, numpy.ndarray)
    arrK = xp.asanyarray(arrK)
    arrI0 = xp.asanyarray(arrI0)
    
    if enInputCheck: # check
        # unfixable
        if not isInputNumpy and not useCuda:
            raise RuntimeError(f"cupy input must be processed via cuda")
        if arrK.ndim!=2 or arrK.shape[1] not in (2,3): 
            raise RuntimeError(f"`arrK.shape` should be `[nK,nDim]`, got `{arrK.shape}`")
        arrRho:NDArray = xp.linalg.norm(arrK, axis=-1)
        if abs(arrRho.max()-0.5)>0.1: raise UserWarning("k-range: [-0.5,0.5]")
        # fixable
        if arrK.shape[1]==3 and (arrK[:,2]==0).all(): arrK = arrK[:,:2]
        if useCuda: arrK = arrK.astype("float32")
    
    # basic parameter
    nPix = int(nPix)
    nK, nAx = arrK.shape
    complex = xp.complex64
    float = xp.float32
    scomplex = "complex64"
    
    if fDbgInfo: print(f"# input check: {time() - t0:.3f}s"); t0 = time()
    
    # data initialize
    arrDcf = xp.ones((nK,), dtype=complex)
    
    # radial DCF
    arrRho = xp.sum(arrK**2, axis=-1, dtype=complex)
    xp.sqrt(arrRho, out=arrRho)
    arrDcf *= (arrRho+1/nPix)**(nAx-1)
    
    # 1D DCF
    if arrI0 is not None:
        arrDcf1D = xp.empty((nK,), dtype=complex)
        arrDcf1D[:-1] = xp.sqrt(xp.sum(xp.diff(arrK, axis=0)**2, axis=-1)) # this step takes 1.2s?
        arrDcf1D[-1] = arrDcf1D[-2]
        arrDcf1D[arrI0[1:]-1] = arrDcf1D[arrI0[1:]-2] # fix the error at seam of two trajectories
        arrDcf *= arrDcf1D
    
    # return arrDcf # see how initial DCF be like
    
    if fDbgInfo: print(f"# data initialize: {time() - t0:.3f}s"); t0 = time()

    # grid of rho
    coords = xp.ogrid[tuple(slice(0, 1, nPix*1j) for _ in range(nAx))]
    arrGridRho = xp.sqrt(py_sum(c.astype(float)**2 for c in coords))
    if fDbgInfo: print(f"# grid of rho: {time() - t0:.3f}s"); t0 = time()

    # Nd window
    if sWind=="poly": arrWindNd = 1 - arrGridRho.clip(0,1)**(2.0 if pShape is None else pShape) # 2.4
    elif sWind=="cos": arrWindNd = xp.cos(arrGridRho*pi/2).clip(0,1)**(1.0 if pShape is None else pShape) # 0.7
    elif sWind=="es": beta=2.0 if pShape is None else pShape; arrWindNd = xp.exp(beta*xp.sqrt(1-arrGridRho.clip(0,1)**2))/xp.exp(beta)
    else: raise NotImplementedError("")
    
    arrWindNd[arrGridRho>1]=0
    del arrGridRho
    
    for iAx in range(nAx):
        tupSli = tuple(0 if iAx==_iAx else slice(None) for _iAx in range(nAx))
        xp.sqrt(arrWindNd[tupSli], out=arrWindNd[tupSli])
    if fDbgInfo: print(f"# Nd window: {time() - t0:.3f}s"); t0 = time()
    
    # deconvolve
    nufftpara = {"upsampfac":1.25} if useCuda else {"debug":0, "spread_debug":0, "showwarn":0, "upsampfac":1.25, "nthreads":cpu_count(logical=True), "spread_sort":1, "fftw":64}
    fn = cufinufft if useCuda else finufft
    
    n_modes = tuple(2*nPix-1 for _ in range(nAx))
    arr2PiKT = xp.array(arrK.T, order='C', dtype=float)
    arr2PiKT *= 2*pi
    eps = 1e-3
    
    pNuift = fn.Plan(1, n_modes, eps=eps, dtype=scomplex, **nufftpara)
    pNufft = fn.Plan(2, n_modes, eps=eps, dtype=scomplex, **nufftpara)
    pNuift.setpts(*arr2PiKT)
    pNufft.setpts(*arr2PiKT)
    for i in range(nStep):
        arrPsf = pNuift.execute(arrDcf)
        if fDbgInfo: print(f"# nuift: {time() - t0:.3f}s"); t0 = time()
        
        # suppress alias outside of PSF
        sliNeg = slice(nPix-1,None,-1)
        sliPos = slice(nPix-1,None,1)
        for iCorner in product(range(2), repeat=nAx):
            tupSli = tuple(sliNeg if i else sliPos for i in iCorner)
            arrPsf[tupSli] *= arrWindNd
        
        arrDcfApo = pNufft.execute(arrPsf)
        arrDcf /= arrDcfApo
        if fDbgInfo: print(f"# nufft: {time() - t0:.3f}s"); t0 = time()
    
    if isInputNumpy and xp!=numpy:
        return arrDcf.get()
    else:
        return arrDcf
        


from .Type import OpDir
from .Utility import grid
from numpy import *
from scipy.special import iv
from copy import deepcopy
from scipy.special import j1

from matplotlib.pyplot import figure, plot, show

def zwarts_kernel(r_squared, rfp_sq):
    """
    Python implementation of the _poly_sdc_kern_0lobes function from the C code.
    Accepts squared radius to avoid sqrt.
    """
    # Polynomial coefficients from the C code
    poly = [
        -1.1469041640943728E-13, 
         8.5313956268885989E-11, 
         1.3282009203652969E-08, 
        -1.7986635886194154E-05, 
         3.4511129626832091E-05, 
         0.99992359966186584
    ]
    
    # Scaling constants from the C code
    SPECTRAL_LEN = 25600.0
    FOV = 63.0
    
    # The polynomial is only valid up to the radius-fov-product
    if r_squared >= rfp_sq:
        return 0.0

    r = sqrt(r_squared)
    x = SPECTRAL_LEN * r / FOV
    
    # Evaluate the polynomial
    out = poly[5] # Zeroth order
    out += poly[4] * x
    out += poly[3] * x**2
    out += poly[2] * x**3
    out += poly[1] * x**4
    out += poly[0] * x**5
    
    return out if out > 0.0 else 0.0

def pipe(_arrK:ndarray, nPix:int, nIt:int=25, ovKer:int=101, ovKsp=2.1, fOptKer:bool=True) -> ndarray:
    # avoid modifying the passed-in value
    arrK = deepcopy(_arrK)

    # normalize `arrK`
    arrK = arrK*(0.5/abs(arrK).max())

    nK, nDim = arrK.shape

    if fOptKer:
        # if nDim==2:
        #     nPix_Ker = int(nPix*0.00954) # cut off at main-lobe
        #     arrRho = linspace(0, 0.5*nPix_Ker/nPix, 100000)
        #     arrRho[0] += 1e-3
        #     arrHalf1dWind = (nPix*j1(nPix*pi*arrRho)) / (2*arrRho)
        #     arrHalf1dWind **= 2
        #     arrHalf1dWind[0] = (nPix**2 * pi / 4)**2
        # elif nDim==3:
        #     nPix_Ker = int(nPix*0.011175) # cut off at main-lobe
        #     arrRho = linspace(0, 0.5*nPix_Ker/nPix, 500)
        #     arrRho[0] += 1e-3
        #     arrHalf1dWind = (sin(pi*nPix*arrRho) - pi*nPix*arrRho*cos(pi*nPix*arrRho)) / (2 * pi**2 * arrRho**3)
        #     arrHalf1dWind **= 2
        #     arrHalf1dWind[0] = (pi * nPix**3 / 6)**2
            
        # kwart's magic kernel, good but I did't figure out why
        RADIUS_FOV_PRODUCT = 0.96960938
        nPix_Ker = RADIUS_FOV_PRODUCT * 2
        rfp_sq = RADIUS_FOV_PRODUCT**2
        arrR_sq = linspace(0, rfp_sq, 5000)
        arrHalf1dWind = array([zwarts_kernel(r_sq, rfp_sq) for r_sq in arrR_sq])
    
    else: # parameter from Pipe's
        ordBessel = 8
        nPix_Ker = 5
        beta = 16
        arrHalf1dWind = iv(ordBessel, beta*sqrt(1-linspace(0,1,nPix_Ker*100,1)**2))/iv(ordBessel, beta)
    
    # arrHalf1dWind = sqrt(arrHalf1dWind) # Zwart recommend to do this because dual-step Cartesian gridding is used instead of a single non-Cartesian method

    nPix_Ksp = int(nPix*ovKsp)
    nPix_Ker = int(nPix_Ker*ovKsp)

    arrDcf = ones([arrK.shape[0]], dtype=complex128)
    arrKsp = ones([(nPix_Ksp+nPix_Ker//2*2+2) for _ in range(nDim)], dtype=complex128)
    for iIt in range(nIt):
        print(f"{iIt}/{nIt}")
        arrKsp = grid(OpDir.FOR, arrDcf, arrK, arrKsp, nPix_Ksp, nPix_Ker, ovKer, arrHalf1dWind)
        arrDcf /= grid(OpDir.BAC, arrDcf, arrK, arrKsp, nPix_Ksp, nPix_Ker, ovKer, arrHalf1dWind)

    return abs(arrDcf)


from numpy import *
from matplotlib.pyplot import *
from .Type import OpDir
from copy import deepcopy

class Grid:
    def __init__(self):
        self.arrS = None
        self.arrK = None
        self.arrGrid = None
        self.arrHalf1dWind = None
        
        pass

    def init(eOpDir:OpDir, _arrS:ndarray, _arrK:ndarray, _arrGrid:ndarray, nPix_Ksp:int, nPix_Ker:int, ov:int, _arrHalf1dWind:ndarray):
        pass

    def run():
        pass
    
def grid(eOpDir:OpDir, _arrS:ndarray, _arrK:ndarray, _arrGrid:ndarray, nPix_Ksp:int, nPix_Ker:int, ov:int, _arrHalf1dWind:ndarray) -> ndarray:
    """
    # note
    this function accept `arrK` in range [-0.5,0.5), but scaled to [-nPix/2,nPix/2) internally
    """
    arrS = deepcopy(_arrS)
    arrK = deepcopy(_arrK)
    arrGrid = deepcopy(_arrGrid)
    arrHalf1dWind = deepcopy(_arrHalf1dWind)

    if arrGrid.shape[0] < int(nPix_Ksp + nPix_Ker//2*2 + 1): # +1: in case k=0.5
        raise ValueError("size of `arrGrid` is not big enough.")

    if ov%2==0: # ensure oversampling factor is odd
        ov+=1
        raise Warning(f"ov changed to {ov} to ensure odd.")

    nPix_Grid = arrGrid.shape[0]
    nPix_Ksp = int(nPix_Ksp)
    nPix_Ker = int(nPix_Ker) # kernel width
    nPix_KerOv = int(nPix_Ker*ov) # kernel width after ov
    nPix_Wind = arrHalf1dWind.size
    nK, nDim = arrK.shape

    # convert /pix to /fov
    arrK *= nPix_Ksp # -Npix/2 ~ Npix/2

    # generate oversampled kernel
    tupKerK = meshgrid\
    (
        *(linspace(-1, 1, nPix_KerOv, 1) for _ in range(nDim)),
        indexing="ij"
    )[::-1]
    arrKerK = array(tupKerK).transpose(*arange(1,nDim+1), 0)
    arrKerR = sqrt(sum((arrKerK)**2, axis=-1))
    arrKer = interp(arrKerR, linspace(0,1,nPix_Wind,1), arrHalf1dWind)

    # generate idx in kernel when on-grid
    tupKerI0 = meshgrid\
    (
        *(arange(nPix_Ker) for _ in range(nDim)),
        indexing="ij"
    )[::-1]
    arrKerI0 = array(tupKerI0).transpose(*arange(1,nDim+1), 0)
    arrKerI0 *= ov
    arrKerI0 += ov//2

    # grid
    if eOpDir==OpDir.FOR:
        arrGrid[:] = 0 # = zeros([nPix_Grid for _ in range(nDim)], dtype=complex128)
    else:
        arrS[:] = 0

    for iK in range(nK):
        # coord of kernel
        if nPix_Ker%2==0:
            arrK_Ker = floor(arrK[iK,:]) + 0.5
        else:
            arrK_Ker = around(arrK[iK,:])
        
        # index of kernel
        biasIdx = around(ov*(arrK_Ker - arrK[iK,:])).astype(int)
        arrKerI = arrKerI0 + biasIdx

        # calculate range of K
        minK = ceil(arrK[iK,:] - nPix_Ker/2).astype(int64)
        maxK = minK + nPix_Ker - 1

        # perform grid
        if eOpDir==OpDir.FOR:
            arrGrid[*(slice(minK[-iDim-1]+nPix_Grid//2, maxK[-iDim-1]+1+nPix_Grid//2) for iDim in range(nDim))] +=\
            arrS[iK]*arrKer[*(arrKerI[*(slice(None) for _ in range(nDim)), -iDim-1] for iDim in range(nDim))]
        elif eOpDir==OpDir.BAC:
            arrS[iK] += sum\
            (
                arrGrid[*(slice(minK[-iDim-1]+nPix_Grid//2, maxK[-iDim-1]+1+nPix_Grid//2) for iDim in range(nDim))]*\
                arrKer[*(arrKerI[*(slice(None) for _ in range(nDim)), -iDim-1] for iDim in range(nDim))]
            )
        else:
            raise NotImplementedError("")

    if eOpDir==OpDir.FOR:
        return arrGrid
    elif eOpDir==OpDir.BAC:
        return arrS
    else:
        raise NotImplementedError("")



import numpy as np
import finufft

# number of nonuniform points
M = 100

# the nonuniform points
x = 2 * np.pi * np.random.uniform(size=M)
y = 2 * np.pi * np.random.uniform(size=M)
z = 2 * np.pi * np.random.uniform(size=M)

# number of Fourier modes
N1, N2, N3 = 50, 75, 100

# the Fourier mode coefficients
f = (np.random.standard_normal(size=(N1, N2, N3))
     + 1J * np.random.standard_normal(size=(N1, N2, N3)))

# calculate the type-2 NUFFT
c = finufft.nufft3d2(x, y, z, f)


# export OMP_NUM_THREAD=16
psum = sum

from numpy import *
from numpy.linalg import norm
from matplotlib.pyplot import *
from os.path import exists
import slime
import finufft as fn
import mrarbgrad as mag
import mrarbdcf as mad
from sigpy.mri.dcf import pipe_menon_dcf
from time import time
from skimage.metrics import structural_similarity as ssim
from scipy.io import savemat, loadmat
import h5py

gamma = 42.5756e6 # UIH
goldrat = (1+sqrt(5))/2

nPix = 256
fov = 0.5
sLim = 100 * gamma * fov/nPix
gLim = 120e-3 * gamma * fov/nPix
dtGrad = 10e-6
dtADC = 5e-6

sPathRes = "/mnt/d/LProject/DcfBenchmark/resource/"
sMethod = ["Baseline", "Proposed"][1]
fDiagFov = 1

if 0:
    # generate phantom
    sPathPhant2D = f"{sPathRes[:-1]}/arrM0_2d.npy"
    sPathPhant3D = f"{sPathRes[:-1]}/arrM0_3d.npy"
    if exists(sPathPhant2D) and exists(sPathPhant3D): # calculate phantom
        arrM0_2d = load(sPathPhant2D)
        arrM0_2d = asarray(arrM0_2d)
        arrM0_3d = load(sPathPhant3D)
        arrM0_3d = asarray(arrM0_3d)
    else:
        random.seed(0)
        arrM0_2d = slime.genPhan(nDim=2, nPix=nPix)["M0"]
        arrM0_2d = asarray(arrM0_2d, dtype=complex64).squeeze()

        random.seed(0)
        arrM0_3d = slime.genPhan(nDim=3, nPix=nPix)["M0"]
        arrM0_3d = asarray(arrM0_3d, dtype=complex64).squeeze()

        save(sPathPhant2D, arrM0_2d)
        save(sPathPhant3D, arrM0_3d)
else:
    arrM0_3d = load(f"{sPathRes[:-1]}/imgRec.npy"); 
    # arrM0_3d = load(f"{sPathRes[:-1]}/1b197efe-9865-43be-ac24-f237c380513e.npy"); 
    # arrM0_3d = load(f"{sPathRes[:-1]}/2588bfa8-0c97-478c-aa5a-487cc88a590d.npy")
    arrM0_3d = asarray(arrM0_3d, dtype=complex64)
    print("phantom", abs(arrM0_3d.mean()))
    assert all(arrM0_3d.shape==(nPix,)*3)
    arrM0_2d = arrM0_3d[nPix//2,...]

# figure()
# imshow(abs(arrM0_2d), cmap="gray")
# show()
# exit()

for sTraj in ["VdSpiral", "VdSpiral", "Rosette", "Yarnball", "Cones"]:
# for sTraj in ["Rosette"]:
    if sTraj in ["VdSpiral", "Rosette"]:
        nAx = 2
        mag.setGoldAng(1)
    elif sTraj in ["Yarnball", "Cones"]:
        nAx = 3
        mag.setGoldAng(0)
    else:
        raise NotImplementedError("")
        
    ovTraj = sqrt(nAx) if fDiagFov else 1
    gLim = amin([gLim, 1/(dtADC*nPix*ovTraj)]) # gLim*dtADC < 1/nPix

    # generate phantom
    if nAx==2: arrM0 = arrM0_2d.copy()
    if nAx==3: arrM0 = arrM0_3d.copy()
    
    # calculate trajectory
    if 0:
        if sTraj=="VdSpiral": lstArrK0, lstArrGrad = mag.getG_VarDenSpiral(dFov=fov*goldrat*ovTraj, lNPix=nPix*goldrat*ovTraj, dSLim=sLim, dGLim=gLim, dDt=dtGrad, dRhoPhi0=0.5/(8*pi), dRhoPhi1=0.5/(2*pi))
        elif sTraj=="Rosette": lstArrK0, lstArrGrad = mag.getG_Rosette(dFov=fov*goldrat*ovTraj/5, lNPix=nPix*goldrat*ovTraj/5, dSLim=sLim, dGLim=gLim, dDt=dtGrad)
        elif sTraj=="Yarnball": lstArrK0, lstArrGrad = mag.getG_Yarnball(dFov=fov*ovTraj, lNPix=nPix*ovTraj, dSLim=sLim, dGLim=gLim, dDt=dtGrad)
        elif sTraj=="Cones": lstArrK0, lstArrGrad = mag.getG_Cones(dFov=fov*ovTraj, lNPix=nPix*ovTraj, dSLim=sLim, dGLim=gLim, dDt=dtGrad)
        else: raise NotImplementedError("")
        
        lstArrK = []
        for arrK0, arrGrad in zip(lstArrK0, lstArrGrad):
            arrK, _ = mag.cvtGrad2Traj(arrGrad, dtGrad, dtADC)
            arrK += arrK0
            lstArrK.append(arrK[:,:nAx])
        arrK = concatenate(lstArrK, axis=0, dtype=float32)[:,:nAx]
        arrNRO = array([arrK.shape[0] for arrK in lstArrK])
        savemat(f"{sPathRes}arrK_{sTraj}.mat", {"arrK":arrK, "arrNRO":arrNRO, "sTraj":sTraj})
        continue # test
    else:
        dicArchive = loadmat(f"{sPathRes}arrK_{sTraj}.mat")
        arrK = asarray(dicArchive["arrK"]).astype(float32)
        arrNRO = asarray(dicArchive["arrNRO"]).astype(int).squeeze()
        sTraj = str(dicArchive["sTraj"].item())
    arrI0 = zeros((arrNRO.size+1,), dtype=int)
    arrI0[1:] = cumsum(arrNRO)
    lstArrK = [arrK[arrI0[i]:arrI0[i+1]] for i in range(arrI0.size-1)]
    
    if sMethod == "Proposed":
        # propsoed
        mad.setDbgInfo(0)
        mad.setInputCheck(0)
        mad.setNumStep(2)
        
        t0 = time()
        # lstArrDcf = mad.sovDcf(int(nPix), lstArrK, sWind="poly")
        arrDcf = mad.calDcf(int(nPix), arrK, arrI0, sWind="poly")
        tExe = time() - t0
        
        arrDcf = mad.normDcf(arrDcf, nAx)
        print(f"Texe: {tExe:.3f}s")
        # continue
    elif sMethod == "Sigpy":
        # sigpy baseline
        t0 = time()
        arrDcf = pipe_menon_dcf(arrK*nPix, [nPix for _ in range(nAx)], max_iter=40, show_pbar=1).astype(complex64)
        tExe = time() - t0
        print(f"Texe: {tExe:.3f}s")
    elif sMethod == "Baseline":
        # external baseline
        with h5py.File(f"{sPathRes}arrDcf_{sTraj}.mat") as f:
            arrDcf = asarray(f["arrDcf"]).astype(complex64).squeeze()
            tExe = asarray(f["Texe"]).astype(float32).squeeze()
        print(f"Texe: {tExe:.3f}s")
    else:
        raise NotImplementedError("")
        
    # simulate kspace
    arr2PiKT = (2*pi)*arrK.T.copy()
    if nAx==2:
        nufft = fn.nufft2d2
        nuift = fn.nufft2d1
    elif nAx==3:
        nufft = fn.nufft3d2
        nuift = fn.nufft3d1
    else:
        raise NotImplementedError("")

    arrS = nufft(*arr2PiKT, arrM0)

    # calculate PSF and M0_Recon
    arrM0Rec = nuift(*arr2PiKT, arrS*arrDcf, tuple(nPix for _ in range(nAx)))
    arrM0Rec *= norm(arrM0.flatten()) / norm(arrM0Rec.flatten()) # normalization suggested by Zwart

    ovPsf = 10
    arrPsf = nuift(*arr2PiKT/ovPsf, arrDcf, tuple(2*nPix for _ in range(nAx))) # must oversamp the PSF so that it can be properly ploted in surface plot view

    # FOV nulling, normalize, evaluate
    coords = ogrid[tuple(slice(-nPix/2, nPix/2, nPix*1j) for _ in range(nAx))]
    arrGridRho = sqrt(psum(float32(c)**2 for c in coords))
    mskFov = ones_like(arrGridRho, dtype=bool) if fDiagFov else arrGridRho < nPix/2 
    
    sNormMethod = "ene"
    arrM0_Norm = mad.normImg(arrM0, sNormMethod, mskFov=mskFov)
    arrM0Rec_Norm = mad.normImg(arrM0Rec, sNormMethod, mskFov=mskFov)
    
    vRange = abs(arrM0_Norm[mskFov]).max()
    metNrmse = sqrt(mean(abs(arrM0Rec_Norm-arrM0_Norm)[mskFov]**2))/vRange
    metSsim = ssim(abs(arrM0_Norm)*mskFov, abs(arrM0Rec_Norm)*mskFov, data_range=vRange)
    
    print(f"{sMethod} {sTraj} NRMSE: {metNrmse:.3f}")
    print(f"SSIM: {metSsim:.3f}")

    # save data
    arrDcf = mad.normDcf(arrDcf, nAx)
    
    savez(f"{sPathRes[:-1]}/{sTraj}_{sMethod}.npz", arrDcf=arrDcf, arrPsf=arrPsf, arrM0Rec=arrM0Rec, tExe=tExe, metSsim=metSsim, metNrmse=metNrmse)
    
    print("")

save(f"{sPathRes[:-1]}/mskFov_{nAx}d.npy", mskFov)

exit()
# plot
wFPage = 4.77
figure(dpi=150, figsize=(wFPage,wFPage/2))
subplot(111)
iK0 = arrI0[len(lstArrK)//16]
iK1 = arrI0[len(lstArrK)//16+1]
plot(abs(arrDcf[iK0:iK1]), ".-", markersize=4, linewidth=1, label=sMethod)
# ylim(-0.1e-8,1.2e-8)
xlabel("index [a.u.]")
ylabel("DCF [a.u.]")
legend(loc=2)
grid(True, "both", "both")
subplots_adjust(0.15,0.2,0.99,0.9)

show()
exit()

figure()
for i in range(nAx):
    plot(lstArrGrad[0][:,i], ".-")
    plot(lstArrGrad[0][:,i], ".-")
    plot(lstArrGrad[0][:,i], ".-")

show()


from numpy import *
from numpy.typing import NDArray
from matplotlib.pyplot import *
import mrarbgrad as mag
import mrarbdcf as mad
from time import time

sTraj = ["VdSpiral", "Rosette", "Yarnball", "Cones"][2]
gamma = 42.5756e6

nPix = 256
fov = 0.5
sLim = 100 * gamma * fov / nPix
gLim = 120e-3 * gamma * fov / nPix
dtGrad = 10e-6
dtADC = 5e-6

if sTraj=="Yarnball":
    nAx = 3; mag.setGoldAng(nAx==2); ovTraj = sqrt(nAx); gLim = amin([gLim, 1/(dtADC*nPix*ovTraj)])
    lstArrK0, lstArrGrad = mag.getG_Yarnball(nAx==3, fov*sqrt(nAx), nPix*sqrt(nAx), sLim, gLim, dtGrad)
elif sTraj=="VdSpiral":
    nAx = 2; mag.setGoldAng(nAx==2); ovTraj = sqrt(nAx); gLim = amin([gLim, 1/(dtADC*nPix*ovTraj)])
    lstArrK0, lstArrGrad = mag.getG_VarDenSpiral(nAx==3, fov*sqrt(nAx), nPix*sqrt(nAx), sLim, gLim, dtGrad)
elif sTraj=="Rosette":
    nAx = 2; mag.setGoldAng(nAx==2); ovTraj = sqrt(nAx); gLim = amin([gLim, 1/(dtADC*nPix*ovTraj)])
    lstArrK0, lstArrGrad = mag.getG_Rosette(nAx==3, fov*sqrt(nAx), nPix*sqrt(nAx), sLim, gLim, dtGrad)
elif sTraj=="Cones":
    nAx = 3; mag.setGoldAng(nAx==2); ovTraj = sqrt(nAx); gLim = amin([gLim, 1/(dtADC*nPix*ovTraj)])
    lstArrK0, lstArrGrad = mag.getG_Cones(nAx==3, fov*sqrt(nAx), nPix*sqrt(nAx), sLim, gLim, dtGrad)

# Convert gradients to k-space coordinates
lstArrK:list[NDArray] = []
for arrK0, arrGrad in zip(lstArrK0, lstArrGrad):
    arrK, _ = mag.cvtGrad2Traj(arrGrad, dtGrad, dtADC)
    arrK += arrK0
    # Keep all 3 dimensions for the 3D case
    lstArrK.append(arrK[:, :nAx])

mad.setDbgInfo(1)
t = time()
lstArrDcf = mad.sovDcf(nPix, lstArrK, sWind="poly")
t = time()-t
print(f"time: {t:.3f}")

# Normalize for 3D (nAx=3)
lstArrDcf = [mad.normDcf(arrDcf, nAx=nAx) for arrDcf in lstArrDcf]

# 4. Visualization
fig = figure(figsize=(12, 5))
idx = len(lstArrK)//4

# 3D Trajectory Plot (Subsampling for performance)
ax = fig.add_subplot(121, projection='3d' if nAx==3 else None)
ax.plot(*lstArrK[idx].T, '.-')
ax.set_title(sTraj)

# DCF Profile Plot
ax = fig.add_subplot(122)
ax.plot(abs(lstArrDcf[idx]), ".-")
ax.set_xlabel("Index")
ax.set_ylabel("DCF")
ax.grid("both")

show()


import h5py
from numpy import *
from numpy.linalg import norm
import matplotlib.pyplot as plt
# Import svds (Singular Value Decomposition for sparse/truncated calculation)
from scipy.sparse.linalg import svds 
from scipy.signal.windows import gaussian

# --- Configuration ---
pathRes = "/mnt/d/LProject/DcfBenchmark/resource/"
dsid = "1b197efe-9865-43be-ac24-f237c380513e"
# dsid = "2588bfa8-0c97-478c-aa5a-487cc88a590d"
FILE_PATH = f'{pathRes[:-1]}/{dsid}.h5'
DATASET_PATH = '/dataset/data'
TARGET_NX = 320 # Target readout size after cropping (to remove oversampling)

with h5py.File(FILE_PATH, 'r') as f:
    print(f"Loading data from {DATASET_PATH}...")
    
    dataset = f[DATASET_PATH][:]
    
    data = dataset['data']
    head = dataset['head'] 
    print("kspace_data_field.dtype", data.dtype)
    
    if len(data) == 0:
        raise ValueError("No k-space data found.")
    
    print("kspace_data_field[0].shape:", data[0].shape)
    print("kspace_data_field[0].dtype:", data[0].dtype)
    
    nCh = head[0]['active_channels']
    nRO = data[0].shape[0]
    nX = nRO // nCh // 2
    
    nAcq = len(data)
    
    vecIdxPhEnc1 = head['idx']['kspace_encode_step_1']
    vecIdxPhEnc2 = head['idx']['kspace_encode_step_2'] 
    vecIdxSlice = head['idx']['slice']
    
    print(f"DEBUG: vecIdxPhEnc1 range: {vecIdxPhEnc1.min()} to {vecIdxPhEnc1.max()}")
    print(f"DEBUG: vecIdxPhEnc2 range: {vecIdxPhEnc2.min()} to {vecIdxPhEnc2.max()}")
    print(f"DEBUG: vecIdxSlice range: {vecIdxSlice.min()} to {vecIdxSlice.max()}")
    
    nY = vecIdxPhEnc1.max() + 1 # Ny
    nZ = vecIdxSlice.max() + 1
    
    print(f"Inferred Dimensions: Channels={nCh}, Readout Raw (Nx_raw)={nX}, Phase Y (Ny)={nY}, Phase Z (Nz)={nZ}")
    
    # read rawdata
    kspMulCh = zeros((nCh, nZ, nY, nX), dtype=complex64)
    for i in range(nAcq):
        iY = vecIdxPhEnc1[i]
        iZ = vecIdxSlice[i]
        arrS = data[i]
        
        kspMulCh[:, iZ, iY, :] = (arrS[0::2] + arrS[1::2]*1j).reshape(nCh, nX)
        
    # I write this jumble during debugging, but it succeed to recovory the dataset 
    kspMulCh = fft.fftn(kspMulCh, axes=(-3,))
    kspMulCh = fft.fftshift(kspMulCh, axes=(-3,))
    kspMulCh = fft.ifftn(kspMulCh, axes=(-3,))
    kspMulCh = fft.ifftshift(kspMulCh, axes=(-3,))
    kspMulCh = fft.fftn(kspMulCh, axes=(-3,))
    kspMulCh = fft.fftshift(kspMulCh, axes=(-3,))
    
    print(f"K-space array shape for reconstruction: {kspMulCh.shape}")
    
    # coil sensitivity map
    sigma = 0.01
    arrGauWin = gaussian(nZ, nZ*sigma)
    arrGauWin = multiply.outer(arrGauWin, gaussian(nY, nY*sigma))
    arrGauWin = multiply.outer(arrGauWin, gaussian(nX, nX*sigma))
    kspCsm = kspMulCh*arrGauWin[None,...]
    
    imgCsm = kspCsm.copy(); tupAxes = (-3,-2,-1)
    imgCsm = fft.ifftshift(imgCsm, axes=tupAxes)
    imgCsm = fft.ifftn(imgCsm, axes=tupAxes)
    imgCsm = fft.fftshift(imgCsm, axes=tupAxes)

    # coil images
    imgMulCh = kspMulCh.copy(); tupAxes = (-3,-2,-1)
    imgMulCh = fft.ifftshift(imgMulCh, axes=tupAxes)
    imgMulCh = fft.ifftn(imgMulCh, axes=tupAxes)
    imgMulCh = fft.fftshift(imgMulCh, axes=tupAxes)
    
    def crop_trans(img:ndarray):
        iY0 = nY//2-nZ//2; iY1 = iY0 + nZ
        iX0 = nX//2-nZ//2; iX1 = iX0 + nZ
        img = img[:,:,iY0:iY1,iX0:iX1] # crop
        img = img.transpose(0,3,2,1)[:,::-1,:,:]
        return img
        
    imgCsm = crop_trans(imgCsm)
    imgMulCh = crop_trans(imgMulCh)
    
    print(f"Full Image domain array shape: {imgMulCh.shape}")

    # reconstructed image
    nCh, nZ, nY, nX = imgMulCh.shape
    
    imgCsm *= sqrt(nZ*nY*nX)/norm(imgCsm.ravel())
    imgMulCh *= sqrt(nZ*nY*nX)/norm(imgMulCh.ravel())
    imgRec = sum(imgCsm.conj()*imgMulCh, axis=0) / (sum(abs(imgCsm)**2, axis=0) + 1e0) # stablizer is necessary here
    # imgRec = sqrt(sum(abs(imgMulCh)**2, axis=0))
    imgRec = squeeze(imgRec)
    
    imgRec /= percentile(abs(imgRec), 95)
    save(f"{pathRes[:-1]}/{dsid}.npy", imgRec) # save
    
    print(f"Final combined image volume shape: {imgRec.shape}")
    
    # plot
    # imgRec = imgCsm[0,...] # test
    sliceZ0 = imgRec[nZ//2, :, :]
    sliceY0 = imgRec[:, nY//2, :]
    sliceX0 = imgRec[:, :, nX//2]
    
    fig, axes = plt.subplots(2, 3, figsize=(15, 5))
    axes = ravel(axes)
    
    # z=0
    axes[0].imshow(abs(sliceZ0), origin="lower", cmap='gray')
    axes[0].set_title(f'z=0')
    axes[0].axis('off')
    
    # y=0
    axes[1].imshow(abs(sliceY0), origin="lower", cmap='gray')
    axes[1].set_title(f'y=0')
    axes[1].axis('off')
    
    # x=0
    axes[2].imshow(abs(sliceX0), origin="lower", cmap='gray')
    axes[2].set_title(f'x=0')
    axes[2].axis('off')
    
    # z=0
    axes[3].imshow(angle(sliceZ0), origin="lower", cmap='hsv', vmin=-pi, vmax=pi)
    axes[3].set_title(f'z=0')
    axes[3].axis('off')
    
    # y=0
    axes[4].imshow(angle(sliceY0), origin="lower", cmap='hsv', vmin=-pi, vmax=pi)
    axes[4].set_title(f'y=0')
    axes[4].axis('off')
    
    # x=0
    axes[5].imshow(angle(sliceX0), origin="lower", cmap='hsv', vmin=-pi, vmax=pi)
    axes[5].set_title(f'x=0')
    axes[5].axis('off')
    
    plt.show()